import json
import re
from collections import OrderedDict

def ordereddict_to_dict(d):
    """
    Helper function to convert an ordered dict to a dict
    This is probably not necessary but slightly cleaner
    """
    return json.loads(json.dumps(d))

def read_ipuzfile(f):
    """
    Read in an ipuz file, return a dictionary of data
    """
    ret = dict()
    # Note that we need to load an OrderedDict
    # as the order of the keys is important
    with open(f, encoding='utf-8') as fid:
        ipuzdata = json.load(fid, object_pairs_hook=OrderedDict)

    # Collect metadata
    # Remove some stuff from the puzzleKind
    kind = ipuzdata.get('kind', ["http://ipuz.org/crossword#1"])[0]
    kind = kind.replace('http://ipuz.org/', '')
    kind = re.sub('\#\d+$', '', kind)
    # We'll need the width and height later
    width = ipuzdata.get('dimensions', {}).get('width')
    height = ipuzdata.get('dimensions', {}).get('height')
    ret['metadata'] = {
      'kind': kind
    , 'author': ipuzdata.get('author')
    , 'title': ipuzdata.get('title')
    , 'copyright': ipuzdata.get('copyright')
    , 'notes': ipuzdata.get('notes')
    , 'width': width
    , 'height': height
    }

    # Get the grid
    # iPuz allows cells to be strings or dicts
    # we convert everything to a dict
    BLOCK = ipuzdata.get('block', '#')
    EMPTY = ipuzdata.get('empty', '0')
    grid = []
    puzzle = ipuzdata['puzzle']
    for y in range(height):
        for x in range(width):
            ipuzcell = puzzle[y][x]
            cell = {'x': x, 'y': y}
            # case 0: this is null
            if ipuzcell is None:
                cell['isEmpty'] = True
            # case 1: we have a string (or int)
            elif isinstance(ipuzcell, (str, int)):
                # cast to string to be safe
                ipuzcell = str(ipuzcell)
                if ipuzcell == BLOCK:
                    cell['isBlock'] = True
                elif ipuzcell is None:
                    cell['isEmpty'] = True
                elif ipuzcell != EMPTY:
                    cell['number'] = str(ipuzcell)
                try:
                    if ipuzcell != BLOCK and ipuzcell is not None:
                        sol = ipuzdata['solution'][y][x]
                        if isinstance(sol, (dict, OrderedDict)):
                            sol = sol['value']
                        cell['solution'] = sol
                except: # no solution
                    pass
            # case 2: we have a dictionary
            else:
                icell = ipuzcell.get('cell', EMPTY)
                if icell == BLOCK:
                    cell['isBlock'] = True
                elif icell is None:
                    cell['isEmpty'] = True
                elif icell != EMPTY:
                    cell['number'] = str(icell)
                cell['style'] = ordereddict_to_dict(ipuzcell.get('style', {}))
                if ipuzcell.get('value'):
                    cell['value'] = ipuzcell.get('value')
                try:
                    if icell != BLOCK and icell is not None:
                        # we pull the solution value from the "solution"
                        # this can either be a string or a dictionary
                        sol = ipuzdata['solution'][y][x]
                        if isinstance(sol, (dict, OrderedDict)):
                            sol = sol['value']
                        cell['solution'] = sol
                except: # no solution
                    pass
                #END try
            #END if/else
            grid.append(cell)
        #END for x
    #END for y
    ret['grid'] = grid

    ## Clues ##
    # Clues don't always come with explicit cell locations, which is unfortunate
    # but we'll handle that in post, so to speak
    ret_clues = []
    # The way clues are set up, it can either be a list or a dictionary
    for title, clues in ipuzdata.get('clues', {}).items():
        #[ {'title': 'Across', 'clues': [...], 'title': 'Down', 'clues': [...]} ]
        thisClues = []
        for clue1 in clues:
            if isinstance(clue1, list):
                # Indicate in the metadata that we are not given explicit cells
                ret['metadata']['noClueCells'] = True
                number, clue = clue1
                number = str(number)
                thisClues.append({'number': number, 'clue': clue})
            else:
                number = str(clue1.get('number', ''))
                clue = clue1.get('clue', '')
                if 'cells' in clue1.keys():
                    # cells are 1-indexed unfortunately
                    cells1 = clue1['cells']
                    cells = []
                    for cell in cells1:
                        cells.append([cell[0]-1, cell[1]-1])
                    thisClues.append({'number': number, 'clue': clue, 'cells': cells})
                else:
                    # if no clue cells we'll have to infer them
                    ret['metadata']['noClueCells'] = True
                    thisClues.append({'number': number, 'clue': clue})
            #END if/else
        #END for clue1
        ret_clues.append({'title': title, 'clues': thisClues})
    #END for title/clues
    ## Hack for CrossFire-exported iPuz files ##
    if len(ret_clues) == 2:
        if ret_clues[0]['title'].lower() == 'down' and ret_clues[1]['title'].lower() == 'across':
            ret_clues = [ret_clues[1], ret_clues[0]]
    #END hack
    ret['clues'] = ret_clues
    return ret



import json
import re
from collections import OrderedDict, defaultdict
import xml.etree.ElementTree as ET
import zipfile
from lxml import etree

# via https://stackoverflow.com/a/51972010
def cleanup_namespaces(input_xml):
    """Remove the stupid namespaces"""
    root = etree.fromstring(input_xml)

    # Iterate through all XML elements
    for elem in root.getiterator():
        # Skip comments and processing instructions,
        # because they do not have names
        if not (
            isinstance(elem, etree._Comment)
            or isinstance(elem, etree._ProcessingInstruction)
        ):
            # Remove a namespace URI in the element's name
            elem.tag = etree.QName(elem).localname

    # Remove unused namespace declarations
    etree.cleanup_namespaces(root)
    return etree.tostring(root).decode('utf-8')

# courtesy of https://stackoverflow.com/a/32842402
def etree_to_ordereddict(t):
    d = OrderedDict()
    t_tag = t.tag
    d[t_tag] = OrderedDict() if t.attrib else None
    children = list(t)
    if children and t_tag != 'clue':
        dd = OrderedDict()
        for dc in map(etree_to_ordereddict, children):
            for k1, v in dc.items():
                if k1 not in dd:
                    dd[k1] = list()
                dd[k1].append(v)
        d = OrderedDict()
        d[t_tag] = OrderedDict()
        for k1, v in dd.items():
            if len(v) == 1:
                d[t_tag][k1] = v[0]
            else:
                d[t_tag][k1] = v
    if t.attrib:
        d[t_tag].update(('@' + k, v) for k, v in t.attrib.items())
    if t.text:
        text = t.text.strip()
        if children or t.attrib:
            if text:
                d[t_tag]['#text'] = text
        else:
            d[t_tag] = text
    elif t_tag == 'clue':
        d[t_tag]['#text'] = ''.join([ET.tostring(_).decode('utf-8') for _ in list(t)])
    return d


def read_jpzfile(f):
    """
    Read in a JPZ file, return a dictionary of data
    """
    ret = dict()
    # Try to open as a zip file
    try:
        with zipfile.ZipFile(f, 'r') as myzip:
            this_file = myzip.namelist()[0]
            with myzip.open(this_file) as fid:
                xml = fid.read()
    except zipfile.BadZipFile:
        with open(f, 'rb') as fid:
            xml = fid.read()
    tree = ET.XML(cleanup_namespaces(xml))
    jpzdata = etree_to_ordereddict(tree)
    # Take the root node (whatever it is)
    jpzdata = jpzdata[list(jpzdata.keys())[0]]
    jpzdata = jpzdata['rectangular-puzzle']

    CROSSWORD_TYPES = ['crossword', 'coded', 'acrostic']
    crossword_type = 'crossword'
    for ct in CROSSWORD_TYPES:
        if ct in jpzdata.keys():
            crossword_type = ct
            break

    # Collect metadata
    kind = crossword_type
    metadata = jpzdata['metadata']
    ret['metadata'] = {
      'kind': kind
    , 'author': metadata.get('creator')
    , 'title': metadata.get('title')
    , 'copyright': metadata.get('copyright')
    , 'notes': metadata.get('description')
    }

    puzzle = jpzdata[crossword_type]
    grid1 = puzzle['grid']
    width = int(grid1['@width'])
    height = int(grid1['@height'])
    ret['metadata']['width'] = width
    ret['metadata']['height'] = height

    # Get the grid
    grid = []
    for c in grid1['cell']:
        y = int(c['@y'])
        x = int(c['@x'])
        cell = {'x': x-1, 'y': y-1}
        value = c.get('@solve-state')
        solution = c.get('@solution')
        number = c.get('@number')
        # if there's a hint, we show the letter
        if c.get('@hint'):
            value = solution

        # TODO: what do we do with "clue" cells?
        if c.get('@type') == 'clue':
            value = solution

        if value:
            cell['value'] = value
        if solution:
            cell['solution'] = solution
        if number:
            cell['number'] = number

        # black squares
        if c.get('@type') == 'block':
            cell['isBlock'] = True
        elif c.get('@type') == 'void':
            cell['isEmpty'] = True
        ## STYLE ##
        # TODO: lots of possibilities for style
        # for now, just focus on a few
        style = {}
        # circle
        if c.get('@background-shape') == 'circle':
             style["shapebg"] = "circle"
        # color
        if c.get('@background-color'):
            style['color'] = c.get('@background-color').replace('#', '')
        # bars
        bar_string = ''
        for letter, side in {'T': 'top', 'B': 'bottom', 'L': 'left', 'R': 'right'}.items():
            if c.get(f'@{side}-bar'):
                bar_string += letter
        if bar_string:
            style['barred'] = bar_string
        # top right numbers
        if c.get('@top-right-number'):
            style['mark'] = {"TR": c.get('@top-right-number')}

        cell['style'] = style

        grid.append(cell)
    ret['grid'] = grid

    ## Clues ##
    # in a jpz, "clues" are separate from "words"
    # so we'll have to handle both

    # Words first
    # helper function to get cell values from "x" and "y" strings
    def cells_from_xy(x, y):
        word_cells = []
        split_x = x.split('-')
        split_y = y.split('-')
        if len(split_x) > 1:
            x_from, x_to = map(int, split_x)
            y1 = int(split_y[0])
            step = 1
            if x_to < x_from:
                step = -1
            for k in range(x_from, x_to + step, step):
                word_cells.append([k-1, y1-1])
        elif len(split_y) > 1:
            y_from, y_to = map(int, split_y)
            x1 = int(split_x[0])
            step = 1
            if y_to < y_from:
                step = -1
            for k in range(y_from, y_to + step, step):
                word_cells.append([x1-1, k-1])
        else:
            word_cells.append([int(split_x[0])-1, int(split_y[0])-1])
        return word_cells

    words = dict()
    for w in puzzle['word']:
        _id = w['@id']
        x = w.get('@x')
        y = w.get('@y')
        cells = []
        if x and y:
            cells = cells_from_xy(x, y)
        else:
            wc = w.get('cells')
            for xy in wc:
                _x, _y = xy.get('@x'), xy.get('@y')
                new_cells = cells_from_xy(_x, _y)
                cells.extend(new_cells)
        words[_id] = cells

    # Now clues for real
    clues = []
    # no clues in a coded crossword
    if crossword_type != 'coded':
        clues1 = puzzle['clues']
        if type(clues1) != list:
            clues1 = [clues1]
        for clue_list in clues1:
            clue_el_title = clue_list.get('title')
            if isinstance(clue_el_title, OrderedDict):
                clue_el_title = list(clue_el_title.values())[0]
            clue_el_clues = clue_list.get('clue', [])
            this_clues = {'title': clue_el_title, 'clues': []}
            for c in clue_el_clues:
                number = c.get('@number')
                word_id = c.get('@word')
                clue_text = c.get('#text')
                cells = words[word_id]
                this_clues['clues'].append({'number': number, 'clue': clue_text, 'cells': cells})
            clues.append(this_clues)

    ret['clues'] = clues
    return ret



import json
import re
from collections import OrderedDict, defaultdict
import xml.etree.ElementTree as ET


# courtesy of https://stackoverflow.com/a/32842402
def etree_to_ordereddict(t):
    d = OrderedDict()
    d[t.tag] = OrderedDict() if t.attrib else None
    children = list(t)
    if children:
        dd = OrderedDict()
        for dc in map(etree_to_ordereddict, children):
            for k, v in dc.items():
                if k not in dd:
                    dd[k] = list()
                dd[k].append(v)
        d = OrderedDict()
        d[t.tag] = OrderedDict()
        for k, v in dd.items():
            if len(v) == 1:
                d[t.tag][k] = v[0]
            else:
                d[t.tag][k] = v
    if t.attrib:
        d[t.tag].update(('@' + k, v) for k, v in t.attrib.items())
    if t.text:
        text = t.text.strip()
        if children or t.attrib:
            if text:
                d[t.tag]['#text'] = text
        else:
            d[t.tag] = text
    return d


def read_cfpfile(f):
    """
    Read in a CFP file, return a dictionary of data
    """
    ret = dict()
    with open(f, 'r') as fid:
        xml = fid.read()
    tree = ET.XML(xml)
    cfpdata = etree_to_ordereddict(tree)
    cfpdata = cfpdata['CROSSFIRE']

    # Collect metadata
    kind = "crossword"
    # We'll need the width and height later
    grid1 = cfpdata['GRID']
    width = int(grid1['@width'])
    grid_text = grid1['#text']
    soln_arr = grid_text.split('\n')
    height = len(soln_arr)
    ret['metadata'] = {
      'kind': kind
    , 'author': cfpdata.get('AUTHOR')
    , 'title': cfpdata.get('TITLE')
    , 'copyright': cfpdata.get('COPYRIGHT')
    , 'notes': cfpdata.get('NOTES')
    , 'width': width
    , 'height': height
    , 'noClueCells': True
    }

    # Read in rebus info, if available
    rebus1 = cfpdata.get('REBUSES', {})
    rebus = dict()
    for _, v in rebus1.items():
        rebus[v['@input']] = v['@letters']

    # Circle info, if available
    circles1 = cfpdata.get('CIRCLES', '-1')
    circles = set(map(int, circles1.split(',')))


    # Get the grid
    grid = []
    for i, letter in enumerate(grid_text.replace('\n', '')):
        y = i // width
        x = i % width
        cell = {'x': x, 'y': y, 'value': None}
        cell_value = rebus.get(letter, letter)
        # black squares
        if cell_value == '.':
            cell_value = None
            cell['isBlock'] = True
        cell['solution'] = cell_value
        # circles
        style = {}
        if i in circles:
            cell['style'] = {"shapebg": "circle"}
        grid.append(cell)
    ret['grid'] = grid

    ## Clues ##
    ret_clues = [{'title': 'Across', 'clues': []}, {'title': 'Down', 'clues': []}]
    for c in cfpdata.get('WORDS', {}).get('WORD', []):
        # {'number': number, 'clue': clue}
        clue = {'number': c.get('@num', ''), 'clue': c.get('#text', '')}
        this_ix = int(c['@dir'].lower() == 'down')
        ret_clues[this_ix]['clues'].append(clue)
    ret['clues'] = ret_clues
    return ret



"""
Copyright (c) 2010 Alex Dejarnatt. MIT License.
"""

import functools
import operator
import math
import string
import struct
import sys

__title__ = 'puzpy'
__version__ = '0.2.3'
__author__ = 'Alex DeJarnatt'
__author_email__ = 'adejarnatt@gmail.com'
__maintainer__ = 'Simeon Visser'
__maintainer_email__ = 'simeonvisser@gmail.com'
__license__ = 'MIT'
__copyright__ = 'Copyright 2009 Alex DeJarnatt'


PY3 = sys.version_info[0] >= 3

if PY3:
    str = str
    range = range
else:
    str = unicode  # noqa: F821
    range = xrange  # noqa: F821

HEADER_FORMAT = '''<
             H 11s        xH
             Q       4s  2sH
             12s         BBH
             H H '''

HEADER_CKSUM_FORMAT = '<BBH H H '

EXTENSION_HEADER_FORMAT = '< 4s  H H '

MASKSTRING = 'ICHEATED'

ENCODING = 'ISO-8859-1'
ENCODING_UTF8 = 'UTF-8'
ENCODING_ERRORS = 'strict'  # raises an exception for bad chars; change to 'replace' for laxer handling

ACROSSDOWN = b'ACROSS&DOWN'

BLACKSQUARE = '.'
BLACKSQUARE2 = ':'


def enum(**enums):
    return type('Enum', (), enums)


PuzzleType = enum(
    Normal=0x0001,
    Diagramless=0x0401
)

# the following diverges from the documentation
# but works for the files I've tested
SolutionState = enum(
    # solution is available in plaintext
    Unlocked=0x0000,
    # solution is locked (scrambled) with a key
    Locked=0x0004
)

GridMarkup = enum(
    # ordinary grid cell
    Default=0x00,
    # marked incorrect at some point
    PreviouslyIncorrect=0x10,
    # currently showing incorrect
    Incorrect=0x20,
    # user got a hint
    Revealed=0x40,
    # circled
    Circled=0x80
)


# refer to Extensions as Extensions.Rebus, Extensions.Markup
Extensions = enum(
    # grid of rebus indices: 0 for non-rebus;
    # i+1 for key i into RebusSolutions map
    Rebus=b'GRBS',
    # map of rebus solution entries eg 0:HEART;1:DIAMOND;17:CLUB;23:SPADE;
    RebusSolutions=b'RTBL',
    # user's rebus entries
    RebusFill=b'RUSR',
    # timer state: 'a,b' where a is the number of seconds elapsed and
    # b is a boolean (0,1) for whether the timer is running
    Timer=b'LTIM',
    # grid cell markup: previously incorrect: 0x10;
    # currently incorrect: 0x20,
    # hinted: 0x40,
    # circled: 0x80
    Markup=b'GEXT'
)


def read(filename):
    """
    Read a .puz file and return the Puzzle object.
    throws PuzzleFormatError if there's any problem with the file format.
    """
    with open(filename, 'rb') as f:
        return load(f.read())


def load(data):
    """
    Read .puz file data and return the Puzzle object.
    throws PuzzleFormatError if there's any problem with the file format.
    """
    puz = Puzzle()
    puz.load(data)
    return puz


class PuzzleFormatError(Exception):
    """
    Indicates a format error in the .puz file. May be thrown due to
    invalid headers, invalid checksum validation, or other format issues.
    """
    def __init__(self, message=''):
        self.message = message


class Puzzle:
    """Represents a puzzle
    """
    def __init__(self):
        """Initializes a blank puzzle
        """
        self.preamble = b''
        self.postscript = b''
        self.title = ''
        self.author = ''
        self.copyright = ''
        self.width = 0
        self.height = 0
        self.version = b'1.3'
        self.fileversion = b'1.3\0'  # default
        self.encoding = ENCODING
        # these are bytes that might be unused
        self.unk1 = b'\0' * 2
        self.unk2 = b'\0' * 12
        self.scrambled_cksum = 0
        self.fill = ''
        self.solution = ''
        self.clues = []
        self.notes = ''
        self.extensions = {}
        # the folowing is so that we can round-trip values in order:
        self._extensions_order = []
        self.puzzletype = PuzzleType.Normal
        self.solution_state = SolutionState.Unlocked
        self.helpers = {}  # add-ons like Rebus and Markup

    def load(self, data):
        s = PuzzleBuffer(data)

        # advance to start - files may contain some data before the
        # start of the puzzle use the ACROSS&DOWN magic string as a waypoint
        # save the preamble for round-tripping
        if not s.seek_to(ACROSSDOWN, -2):
            raise PuzzleFormatError("Data does not appear to represent a "
                                    "puzzle. Are you sure you didn't intend "
                                    "to use read?")

        self.preamble = s.data[:s.pos]

        puzzle_data = s.unpack(HEADER_FORMAT)
        cksum_gbl = puzzle_data[0]
        # acrossDown = puzzle_data[1]
        cksum_hdr = puzzle_data[2]
        cksum_magic = puzzle_data[3]
        self.fileversion = puzzle_data[4]
        # since we don't know the role of these bytes, just round-trip them
        self.unk1 = puzzle_data[5]
        self.scrambled_cksum = puzzle_data[6]
        self.unk2 = puzzle_data[7]
        self.width = puzzle_data[8]
        self.height = puzzle_data[9]
        numclues = puzzle_data[10]
        self.puzzletype = puzzle_data[11]
        self.solution_state = puzzle_data[12]

        self.version = self.fileversion[:3]
        # Once we have fileversion we can guess the encoding
        self.encoding = ENCODING if self.version_tuple()[0] < 2 else ENCODING_UTF8
        s.encoding = self.encoding

        self.solution = s.read(self.width * self.height).decode(self.encoding)
        self.fill = s.read(self.width * self.height).decode(self.encoding)

        self.title = s.read_string()
        self.author = s.read_string()
        self.copyright = s.read_string()

        self.clues = [s.read_string() for i in range(0, numclues)]
        self.notes = s.read_string()

        ext_cksum = {}
        while s.can_unpack(EXTENSION_HEADER_FORMAT):
            code, length, cksum = s.unpack(EXTENSION_HEADER_FORMAT)
            ext_cksum[code] = cksum
            # extension data is represented as a null-terminated string,
            # but since the data can contain nulls we can't use read_string
            self.extensions[code] = s.read(length)
            s.read(1)  # extensions have a trailing byte
            # save the codes in order for round-tripping
            self._extensions_order.append(code)

        # sometimes there's some extra garbage at
        # the end of the file, usually \r\n
        if s.can_read():
            self.postscript = s.read_to_end()

        if cksum_gbl != self.global_cksum():
            raise PuzzleFormatError('global checksum does not match')
        if cksum_hdr != self.header_cksum():
            raise PuzzleFormatError('header checksum does not match')
        if cksum_magic != self.magic_cksum():
            raise PuzzleFormatError('magic checksum does not match')
        for code, cksum_ext in ext_cksum.items():
            if cksum_ext != data_cksum(self.extensions[code]):
                raise PuzzleFormatError(
                    'extension %s checksum does not match' % code
                )

    def save(self, filename):
        puzzle_bytes = self.tobytes()
        with open(filename, 'wb') as f:
            f.write(puzzle_bytes)

    def tobytes(self):
        s = PuzzleBuffer(encoding=self.encoding)
        # commit any changes from helpers
        for h in self.helpers.values():
            if 'save' in dir(h):
                h.save()

        # include any preamble text we might have found on read
        s.write(self.preamble)

        s.pack(HEADER_FORMAT,
               self.global_cksum(), ACROSSDOWN,
               self.header_cksum(), self.magic_cksum(),
               self.fileversion, self.unk1, self.scrambled_cksum,
               self.unk2, self.width, self.height,
               len(self.clues), self.puzzletype, self.solution_state)

        s.write(self.encode(self.solution))
        s.write(self.encode(self.fill))

        s.write_string(self.title)
        s.write_string(self.author)
        s.write_string(self.copyright)

        for clue in self.clues:
            s.write_string(clue)

        s.write_string(self.notes)

        # do a bit of extra work here to ensure extensions round-trip in the
        # order they were read. this makes verification easier. But allow
        # for the possibility that extensions were added or removed from
        # self.extensions
        ext = dict(self.extensions)
        for code in self._extensions_order:
            data = ext.pop(code, None)
            if data:
                s.pack(EXTENSION_HEADER_FORMAT, code,
                       len(data), data_cksum(data))
                s.write(data + b'\0')

        for code, data in ext.items():
            s.pack(EXTENSION_HEADER_FORMAT, code, len(data), data_cksum(data))
            s.write(data + b'\0')

        # postscript is initialized, read, and stored as bytes. In case it is
        # overwritten as a string, this try/except converts it back.
        try:
            postscript_bytes = self.encode(self.postscript)
        except AttributeError:
            postscript_bytes = self.postscript

        s.write(postscript_bytes)

        return s.tobytes()

    def encode(self, s):
        return s.encode(self.encoding, ENCODING_ERRORS)

    def encode_zstring(self, s):
        return self.encode(s) + b'\0'

    def version_tuple(self):
        return tuple(map(int, self.version.split(b'.')))

    def has_rebus(self):
        return self.rebus().has_rebus()

    def rebus(self):
        return self.helpers.setdefault('rebus', Rebus(self))

    def has_markup(self):
        return self.markup().has_markup()

    def markup(self):
        return self.helpers.setdefault('markup', Markup(self))

    def clue_numbering(self):
        numbering = DefaultClueNumbering(self.fill, self.clues, self.width, self.height)
        return self.helpers.setdefault('clues', numbering)

    def blacksquare(self):
        return BLACKSQUARE2 if self.puzzletype == PuzzleType.Diagramless else BLACKSQUARE

    def is_solution_locked(self):
        return bool(self.solution_state != SolutionState.Unlocked)

    def unlock_solution(self, key):
        if self.is_solution_locked():
            unscrambled = unscramble_solution(self.solution, self.width, self.height, key,
                                              ignore_chars=self.blacksquare())
            if not self.check_answers(unscrambled):
                return False

            # clear the scrambled bit and cksum
            self.solution = unscrambled
            self.scrambled_cksum = 0
            self.solution_state = SolutionState.Unlocked

        return True

    def lock_solution(self, key):
        if not self.is_solution_locked():
            # set the scrambled bit and cksum
            self.scrambled_cksum = scrambled_cksum(self.solution, self.width, self.height,
                                                   ignore_chars=self.blacksquare(), encoding=self.encoding)
            self.solution_state = SolutionState.Locked
            scrambled = scramble_solution(self.solution, self.width, self.height, key,
                                          ignore_chars=self.blacksquare())
            self.solution = scrambled

    def check_answers(self, fill):
        if self.is_solution_locked():
            scrambled = scrambled_cksum(fill, self.width, self.height,
                                        ignore_chars=self.blacksquare(), encoding=self.encoding)
            return scrambled == self.scrambled_cksum
        else:
            return fill == self.solution

    def header_cksum(self, cksum=0):
        return data_cksum(struct.pack(HEADER_CKSUM_FORMAT,
                          self.width, self.height, len(self.clues),
                          self.puzzletype, self.solution_state), cksum)

    def text_cksum(self, cksum=0):
        # for the checksum to work these fields must be added in order with
        # null termination, followed by all non-empty clues without null
        # termination, followed by notes (but only for version >= 1.3)
        if self.title:
            cksum = data_cksum(self.encode_zstring(self.title), cksum)
        if self.author:
            cksum = data_cksum(self.encode_zstring(self.author), cksum)
        if self.copyright:
            cksum = data_cksum(self.encode_zstring(self.copyright), cksum)

        for clue in self.clues:
            if clue:
                cksum = data_cksum(self.encode(clue), cksum)

        # notes included in global cksum starting v1.3 of format
        if self.version_tuple() >= (1, 3) and self.notes:
            cksum = data_cksum(self.encode_zstring(self.notes), cksum)

        return cksum

    def global_cksum(self):
        cksum = self.header_cksum()
        cksum = data_cksum(self.encode(self.solution), cksum)
        cksum = data_cksum(self.encode(self.fill), cksum)
        cksum = self.text_cksum(cksum)
        # extensions do not seem to be included in global cksum
        return cksum

    def magic_cksum(self):
        cksums = [
            self.header_cksum(),
            data_cksum(self.encode(self.solution)),
            data_cksum(self.encode(self.fill)),
            self.text_cksum()
        ]

        cksum_magic = 0
        for (i, cksum) in enumerate(reversed(cksums)):
            cksum_magic <<= 8
            cksum_magic |= (
                ord(MASKSTRING[len(cksums) - i - 1]) ^ (cksum & 0x00ff)
            )
            cksum_magic |= (
                (ord(MASKSTRING[len(cksums) - i - 1 + 4]) ^ (cksum >> 8)) << 32
            )

        return cksum_magic


class PuzzleBuffer:
    """PuzzleBuffer class
    wraps a data buffer ('' or []) and provides .puz-specific methods for
    reading and writing data
    """
    def __init__(self, data=None, encoding=ENCODING):
        self.data = data or []
        self.encoding = encoding
        self.pos = 0

    def can_read(self, n_bytes=1):
        return self.pos + n_bytes <= len(self.data)

    def length(self):
        return len(self.data)

    def read(self, n_bytes):
        start = self.pos
        self.pos += n_bytes
        return self.data[start:self.pos]

    def read_to_end(self):
        start = self.pos
        self.pos = self.length()
        return self.data[start:self.pos]

    def read_string(self):
        return self.read_until(b'\0')

    def read_until(self, c):
        start = self.pos
        self.seek_to(c, 1)  # read past
        return str(self.data[start:self.pos-1], self.encoding)

    def seek_to(self, s, offset=0):
        try:
            self.pos = self.data.index(s, self.pos) + offset
            return True
        except ValueError:
            # s not found, advance to end
            self.pos = self.length()
            return False

    def write(self, s):
        self.data.append(s)

    def write_string(self, s):
        s = s or ''
        self.data.append(s.encode(self.encoding, ENCODING_ERRORS) + b'\0')

    def pack(self, struct_format, *values):
        self.data.append(struct.pack(struct_format, *values))

    def can_unpack(self, struct_format):
        return self.can_read(struct.calcsize(struct_format))

    def unpack(self, struct_format):
        start = self.pos
        try:
            res = struct.unpack_from(struct_format, self.data, self.pos)
            self.pos += struct.calcsize(struct_format)
            return res
        except struct.error:
            message = 'could not unpack values at {} for format {}'.format(
                start, struct_format
            )
            raise PuzzleFormatError(message)

    def tobytes(self):
        return b''.join(self.data)


# clue numbering helper

class DefaultClueNumbering:
    def __init__(self, grid, clues, width, height):
        self.grid = grid
        self.clues = clues
        self.width = width
        self.height = height

        # compute across & down
        a = []
        d = []
        c = 0
        n = 1
        for i in range(0, len(grid)):
            if not is_blacksquare(grid[i]):
                lastc = c
                is_across = self.col(i) == 0 or is_blacksquare(grid[i - 1])
                if is_across and self.len_across(i) > 1:
                    a.append({
                        'num': n,
                        'clue': clues[c],
                        'clue_index': c,
                        'cell': i,
                        'len': self.len_across(i)
                    })
                    c += 1
                is_down = self.row(i) == 0 or is_blacksquare(grid[i - width])
                if is_down and self.len_down(i) > 1:
                    d.append({
                        'num': n,
                        'clue': clues[c],
                        'clue_index': c,
                        'cell': i,
                        'len': self.len_down(i)
                    })
                    c += 1
                if c > lastc:
                    n += 1

        self.across = a
        self.down = d

    def col(self, index):
        return index % self.width

    def row(self, index):
        return int(math.floor(index / self.width))

    def len_across(self, index):
        for c in range(0, self.width - self.col(index)):
            if is_blacksquare(self.grid[index + c]):
                return c
        return c + 1

    def len_down(self, index):
        for c in range(0, self.height - self.row(index)):
            if is_blacksquare(self.grid[index + c*self.width]):
                return c
        return c + 1


class Rebus:
    def __init__(self, puzzle):
        self.puzzle = puzzle
        # parse rebus data
        rebus_data = self.puzzle.extensions.get(Extensions.Rebus, b'')
        self.table = parse_bytes(rebus_data)
        r_sol_data = self.puzzle.extensions.get(Extensions.RebusSolutions, b'')
        solutions_str = r_sol_data.decode(puzzle.encoding)
        fill_data = self.puzzle.extensions.get(Extensions.RebusFill, b'')
        fill_str = fill_data.decode(puzzle.encoding)
        self.solutions = dict(
            (int(item[0]), item[1])
            for item in parse_dict(solutions_str).items()
        )
        self.fill = dict(
            (int(item[0]), item[1])
            for item in parse_dict(fill_str).items()
        )

    def has_rebus(self):
        return Extensions.Rebus in self.puzzle.extensions

    def is_rebus_square(self, index):
        return bool(self.table[index])

    def get_rebus_squares(self):
        return [i for i, b in enumerate(self.table) if b]

    def get_rebus_solution(self, index):
        if self.is_rebus_square(index):
            return self.solutions[self.table[index] - 1]
        return None

    def get_rebus_fill(self, index):
        if self.is_rebus_square(index):
            return self.fill[self.table[index] - 1]
        return None

    def set_rebus_fill(self, index, value):
        if self.is_rebus_square(index):
            self.fill[self.table[index] - 1] = value

    def save(self):
        if self.has_rebus():
            # commit changes back to puzzle.extensions
            self.puzzle.extensions[Extensions.Rebus] = pack_bytes(self.table)
            rebus_solutions = self.puzzle.encode(dict_to_string(self.solutions))
            self.puzzle.extensions[Extensions.RebusSolutions] = rebus_solutions
            rebus_fill = self.puzzle.encode(dict_to_string(self.fill))
            self.puzzle.extensions[Extensions.RebusFill] = rebus_fill


class Markup:
    def __init__(self, puzzle):
        self.puzzle = puzzle
        # parse markup data
        markup_data = self.puzzle.extensions.get(Extensions.Markup, b'')
        self.markup = parse_bytes(markup_data)

    def has_markup(self):
        return any(bool(b) for b in self.markup)

    def get_markup_squares(self):
        return [i for i, b in enumerate(self.markup) if b]

    def is_markup_square(self, index):
        return bool(self.markup[index])

    def save(self):
        if self.has_markup():
            self.puzzle.extensions[Extensions.Markup] = pack_bytes(self.markup)


# helper functions for cksums and scrambling
def data_cksum(data, cksum=0):
    for b in data:
        if isinstance(b, bytes):
            b = ord(b)
        # right-shift one with wrap-around
        lowbit = (cksum & 0x0001)
        cksum = (cksum >> 1)
        if lowbit:
            cksum = (cksum | 0x8000)

        # then add in the data and clear any carried bit past 16
        cksum = (cksum + b) & 0xffff

    return cksum


def replace_chars(s, chars, replacement=''):
    for ch in chars:
        s = s.replace(ch, replacement)
    return s


def scramble_solution(solution, width, height, key, ignore_chars=BLACKSQUARE):
    sq = square(solution, width, height)
    data = restore(sq, scramble_string(replace_chars(sq, ignore_chars), key))
    return square(data, height, width)


def scramble_string(s, key):
    """
    s is the puzzle's solution in column-major order, omitting black squares:
    i.e. if the puzzle is:
        C A T
        # # A
        # # R
    solution is CATAR


    Key is a 4-digit number in the range 1000 <= key <= 9999

    """
    key = key_digits(key)
    for k in key:          # foreach digit in the key
        s = shift(s, key)  # for each char by each digit in the key in sequence
        s = s[k:] + s[:k]  # cut the sequence around the key digit
        s = shuffle(s)     # do a 1:1 shuffle of the 'deck'

    return s


def unscramble_solution(scrambled, width, height, key, ignore_chars=BLACKSQUARE):
    # width and height are reversed here
    sq = square(scrambled, width, height)
    data = restore(sq, unscramble_string(replace_chars(sq, ignore_chars), key))
    return square(data, height, width)


def unscramble_string(s, key):
    key = key_digits(key)
    l = len(s)  # noqa: E741
    for k in key[::-1]:
        s = unshuffle(s)
        s = s[l-k:] + s[:l-k]
        s = unshift(s, key)

    return s


def scrambled_cksum(scrambled, width, height, ignore_chars=BLACKSQUARE, encoding=ENCODING):
    data = replace_chars(square(scrambled, width, height), ignore_chars)
    return data_cksum(data.encode(encoding, ENCODING_ERRORS))


def key_digits(key):
    return [int(c) for c in str(key).zfill(4)]


def square(data, w, h):
    aa = [data[i:i+w] for i in range(0, len(data), w)]
    return ''.join(
        [''.join([aa[r][c] for r in range(0, h)]) for c in range(0, w)]
    )


def shift(s, key):
    atoz = string.ascii_uppercase
    return ''.join(
        atoz[(atoz.index(c) + key[i % len(key)]) % len(atoz)]
        for i, c in enumerate(s)
    )


def unshift(s, key):
    return shift(s, [-k for k in key])


def shuffle(s):
    mid = int(math.floor(len(s) / 2))
    items = functools.reduce(operator.add, zip(s[mid:], s[:mid]))
    return ''.join(items) + (s[-1] if len(s) % 2 else '')


def unshuffle(s):
    return s[1::2] + s[::2]


def restore(s, t):
    """
    s is the source string, it can contain '.'
    t is the target, it's smaller than s by the number of '.'s in s

    Each char in s is replaced by the corresponding
    char in t, jumping over '.'s in s.

    >>> restore('ABC.DEF', 'XYZABC')
    'XYZ.ABC'
    """
    t = (c for c in t)
    return ''.join(next(t) if not is_blacksquare(c) else c for c in s)


def is_blacksquare(c):
    if isinstance(c, int):
        c = chr(c)
    return c in [BLACKSQUARE, BLACKSQUARE2]


#
# functions for parsing / serializing primitives
#


def parse_bytes(s):
    return list(struct.unpack('B' * len(s), s))


def pack_bytes(a):
    return struct.pack('B' * len(a), *a)


# dict string format is k1:v1;k2:v2;...;kn:vn;
# (for whatever reason there's a trailing ';')
def parse_dict(s):
    return dict(p.split(':') for p in s.split(';') if ':' in p)


def dict_to_string(d):
    return ';'.join(':'.join(map(str, [k, v])) for k, v in d.items()) + ';'



import json
import base64

def read_amuselabs_data(s):
    """
    Read in an amuselabs string, return a dictionary of data
    """
    # Data might be base64'd or not
    try:
        data = json.loads(s)
    except json.JSONDecodeError:
        s1 = base64.b64decode(s)
        data = json.loads(s1)

    ret = {}

    # metadata
    # technically these can be codewords but i've never seen one
    kind = "crossword"
    width, height = data['w'], data['h']
    ret['metadata'] = {
      'width': width
    , 'height': height
    , 'kind': kind
    , 'author': data.get('author')
    , 'title': data.get('title')
    , 'copyright': data.get('copyright')
    , 'noClueCells': True
    # no notepad?
    }

    # grid
    grid = []
    box = data['box']
    cellInfos = data.get('cellInfos', [])
    # Reshape cellInfos to make lookup easier
    markup = {}
    for c in cellInfos:
        markup[(c['x'], c['y'])] = c
    for y in range(height):
        for x in range(width):
            cell = {'x': x, 'y': y, 'value': None}
            if box[x][y] == '\x00':
                cell['isBlock'] = True
            else:
                cell['solution'] = box[x][y]
            style = {}
            if markup.get((x, y)):
                thisMarkup = markup[(x, y)]
                if thisMarkup.get('isCircled'):
                    style['shapebg'] = 'circle'
                if thisMarkup.get('isVoid'):
                    cell['isBlock'] = False
                    cell['isVoid'] = True
                bar_string = ''
                for letter, side in {'B': 'bottom', 'R': 'right'}.items():
                    if thisMarkup.get(f'{side}Wall'):
                        bar_string += letter
                if bar_string:
                    style['barred'] = bar_string
                cell['style'] = style
            grid.append(cell)
    ret['grid'] = grid

    # clues
    placed_words = data['placedWords']
    across_words = [word for word in placed_words if word['acrossNotDown']]
    down_words = [word for word in placed_words if not word['acrossNotDown']]
    # sorting is probably unnecessary
    across_words = sorted(across_words, key=lambda x: (x['y'], x['x']))
    down_words = sorted(down_words, key=lambda x: (x['y'], x['x']))
    across_clues = [{'number': str(x['clueNum']), 'clue': x['clue']['clue']} for x in across_words]
    down_clues = [{'number': str(x['clueNum']), 'clue': x['clue']['clue']} for x in down_words]
    ret['clues'] = [{'title': 'Across', 'clues': across_clues}, {'title': 'Down', 'clues': down_clues}]
    return ret



from setuptools import setup, find_packages

setup(
    name='pypuz',
    version='0.1.1',
    author='Alex Boisvert et al.',
    author_email='crosswordnexus@gmail.com',
    packages=find_packages(),
    #url='http://pypi.python.org/pypi/PackageName/',
    license='LICENSE.txt',
    description='Reading and writing crossword files in Python',
    long_description=open('README.txt').read(),
    install_requires=[],
)



from .file_types import puz, ipuz, cfp, jpz, amuselabs
import json
import itertools
from collections import OrderedDict

# This imports the _module_ unidecode, which converts Unicode strings to
# plain ASCII. The puz format, however, can accept Latin1, which is a larger
# subset. So the second line tells the module to leave codepoints 128-256
# untouched, then we import the _function_ unidecode.
import unidecode
unidecode.Cache[0] = [chr(c) if c > 127 else '' for c in range(256)]
from unidecode import unidecode as unidecode_fxn

CROSSWORD_TYPE = 'crossword'

# Class for crossword metadata
# This is a mostly uninteresting class
class MetaData:
    """
    Metadata for a crossword.
    Mandatory components:
    kind (the puzzle type)
    Optional:
    author, title, copyright, notes, width, height
    """
    def __init__(self, kind):
        self.kind = kind
#END class MetaData

# Class for a crossword cell
class Cell:
    """
    A crossword cell.
    The only necessary components are x and y (the coordinates, 0-indexed).
    Other possible components:
    solution (string)
    value (string -- the currently filled value of the cell)
    number (string -- the number or letter in the top left)
    isBlock (boolean -- set to True if the cell is a black square)
    isEmpty (boolean -- set to True if the cell is a "void")
    style (dictionary -- see the "StyleSpec" section on http://www.ipuz.org/)
    """
    def __init__(self, x, y, solution=None, value=None, number=None, isBlock=None, isEmpty=None, style={}):
        self.x = x
        self.y = y
        self.solution = solution
        self.value = value
        if number:
            self.number = str(number)
        else:
            self.number = None
        self.isBlock = isBlock
        self.isEmpty = isEmpty
        self.style = style

    def __repr__(self):
        return f"Cell({{({self.x}, {self.y}), {self.solution}}})"

class Grid:
    """
    Class for a crossword grid
    """
    # here "cells" is a list of Cell objects
    def __init__(self, cells):
        self.cells = cells
        self.height = max(c.y for c in cells) + 1
        self.width = max(c.x for c in cells) + 1

    def __repr__(self):
        return json.dumps(self.solutionArray())

    def solutionArray(self):
        arr = []
        for y in range(self.height):
            row = []
            for x in range(self.width):
                c = self.cellAt(x, y)
                letter = c.solution
                if c.isBlock:
                    letter = '#'
                if c.isEmpty:
                    letter = '_'
                row.append(letter)
            arr.append(row)
        return arr

    # return the cell at (x,y)
    def cellAt(self, x, y):
        for c in self.cells:
            if c.x == x and c.y == y:
                return c

    # Return the solution at (x, y)
    def letterAt(self, x, y):
        return self.cellAt(x, y).solution

    # Return True if the cell at (x,y) is empty or a block
    def isBlack(self, x, y):
        thisCell = self.cellAt(x, y)
        return (thisCell.isEmpty or thisCell.isBlock)

    # check if we have a black square (or a bar) in a given direction
    def hasBlack(self, x, y, dir):
        mapping_dict = {
          'R': {'xcheck': self.width-1, 'xoffset': 1, 'yoffset': 0, 'dir2': 'L'}
        , 'L': {'xcheck': 0, 'xoffset': -1, 'yoffset': 0, 'dir2': 'R'}
        , 'T': {'ycheck': 0, 'xoffset': 0, 'yoffset': -1, 'dir2': 'B'}
        , 'B': {'ycheck': self.height-1, 'xoffset': 0, 'yoffset': 1, 'dir2': 'T'}
        }
        md = mapping_dict[dir]
        dir2 = md['dir2']
        if x == md.get('xcheck') or y == md.get('ycheck'):
            return True
        elif self.isBlack(x + md['xoffset'], y + md['yoffset']):
            return True
        elif dir in self.cellAt(x, y).style.get('barred', ''):
            return True
        elif dir2 in self.cellAt(x + md['xoffset'], y + md['yoffset']).style.get('barred', ''):
            return True
        return False
    #END hasBlack

    # Whether the coordinates are the starting letter for a word
    # both startAcrossWord and startDownWord have to account for bars
    def startAcrossWord(self, x, y):
        return self.hasBlack(x, y, 'L') and not self.isBlack(x, y) and not self.hasBlack(x, y, 'R')
    def startDownWord(self, x, y):
        return self.hasBlack(x, y, 'T') and not self.isBlack(x, y) and not self.hasBlack(x, y, 'B')

    # Set the default numbers
    def setNumbering(self):
        thisNumber = 1
        for y in range(self.height):
            thisNumbers = []
            for x in range(self.width):
                if self.startAcrossWord(x, y) or self.startDownWord(x, y):
                    self.cellAt(x, y).number = str(thisNumber)
                    thisNumber += 1
                #END if
            #END for x
        #END for y
    #END def gridNumbering

    # Return the across entries (TODO)
    def acrossEntries(self):
        # Set the numbering if it doesn't already exist
        self.setNumbering()
        acrossEntries = {}
        thisNum = None
        for y in range(self.height):
            for x in range(self.width):
                if self.startAcrossWord(x, y):
                    thisNum = self.cellAt(x, y).number
                    if acrossEntries.get(thisNum) is None and thisNum is not None:
                        acrossEntries[thisNum] = {'word': '', 'cells': []}
                if not self.isBlack(x, y) and thisNum is not None:
                    letter = self.letterAt(x, y)
                    acrossEntries[thisNum]['word'] += letter or ''
                    acrossEntries[thisNum]['cells'].append([x, y])
                # end the across entry if we hit an edge
                if self.hasBlack(x, y, 'R'):
                    thisNum = None
        return acrossEntries
    #END acrossEntries()

    # Return the down entries (TODO)
    def downEntries(self):
        # Set the numbering if it doesn't already exist
        self.setNumbering()
        downEntries = {}
        thisNum = None
        for x in range(self.width):
            for y in range(self.height):
                if self.startDownWord(x, y):
                    thisNum = self.cellAt(x, y).number
                    if downEntries.get(thisNum) is None and thisNum is not None:
                        downEntries[thisNum] = {'word': '', 'cells': []}
                if not self.isBlack(x, y) and thisNum is not None:
                    letter = self.letterAt(x, y)
                    downEntries[thisNum]['word'] += letter or ''
                    downEntries[thisNum]['cells'].append([x, y])
                # end the down entry if we hit the bottom
                if self.hasBlack(x, y, 'B'):
                    thisNum = None
        return downEntries
#END class Grid

class Clue:
    """
    The class for an individual clue
    This has three components:
    * clue -- the actual clue
    * number (optional) -- the number of the clue
    * cells -- the coordinates of the cells for the entry corresponding to the clue
    """
    def __init__(self, clue, cells, number=None):
        self.clue = clue
        self.cells = cells
        if number is not None:
            self.number = str(number)
        else:
            self.number = None

    def __repr__(self):
        return self.clue
#END Clue

class Puzzle:
    """
    Class for a crossword
    """
    def __init__(self, metadata=None, grid=None, clues=None):
        # metadata is a MetaData class
        self.metadata = metadata
        # grid is a Grid class
        self.grid = grid
        # clues is just a list of dictionaries, e.g.
        # [ {'title': 'Across', 'clues': [...], 'title': 'Down', 'clues': [...]} ]
        self.clues = clues

    def fromPuz(self, puzFile):
        # Read in the file
        pz = puz.read(puzFile)

        # Set up the metadata
        kind = CROSSWORD_TYPE
        if pz.puzzletype == 1025:
            kind = 'diagramless'
        metadata = MetaData(kind)
        metadata.title = pz.title
        metadata.author = pz.author
        metadata.copyright = pz.copyright
        metadata.notes = pz.notes
        metadata.width = pz.width
        metadata.height = pz.height

        # Create the grid
        cells = []
        i = 0
        for y in range(metadata.height):
            for x in range(metadata.width):
                cell_value, isBlock = pz.solution[i], None
                fill = pz.fill[i]
                if fill in ('-', '.', ':'):
                    fill = None
                # black squares can occasionally be ":" in puz files
                if cell_value in ('.', ':'):
                    cell_value, isBlock = None, True
                # Rebus
                if pz.has_rebus():
                    r = pz.rebus()
                    if i in r.get_rebus_squares():
                        cell_value = r.get_rebus_solution(i)
                # Circles
                style = {}
                if pz.has_markup():
                    m = pz.markup()
                    if i in m.get_markup_squares():
                        style = {"shapebg": "circle"}
                cell = Cell(x, y, solution=cell_value, value=fill, isBlock=isBlock, style=style)
                cells.append(cell)
                i += 1
            #END for x
        #END for y
        grid = Grid(cells)
        # Set the numbering
        grid.setNumbering()

        # clues
        # Get the across and down entries
        adEntries = (grid.acrossEntries(), grid.downEntries())
        numbering = pz.clue_numbering()
        acrossClues, downClues = numbering.across, numbering.down
        allClues = [acrossClues, downClues]
        clues = [ {'title': 'Across', 'clues': []}, {'title': 'Down', 'clues': []} ]
        for i, clueList in enumerate(allClues):
            for c in clueList:
                number = str(c['num'])
                clue = c['clue']
                cells = adEntries[i][number]['cells']
                clue = Clue(clue=clue, cells=cells, number=number)
                clues[i]['clues'].append(clue)

        return Puzzle(metadata=metadata, grid=grid, clues=clues)
    #END fromPuz()
    
    def toPuz(self, filename):
        """
        Write a .puz file.
        Because of limitations of the .puz format, this is lossy at best.
        In rare cases this may result in a nonsense .puz file
        99% of the time this should work.
        
        Many thanks to xword-dl for the bulk of this code.
        """
        pz = puz.Puzzle()
        # Metadata
        for a in ('author', 'title', 'copyright', 'notes'):
            setattr(pz, a, getattr(self.metadata, a, ''))
            
        # Dimensions
        pz.width, pz.height = self.grid.width, self.grid.height
        
        # Fill and solution
        circled = [(c.x, c.y) for c in self.grid.cells if c.style.get('shapebg') == 'circle']
        solution, fill, markup, rebus_board, rebus_index, rebus_table = '', '', b'', [], 0, ''
        
        for row_num in range(self.grid.height):
            for col_num in range(self.grid.width):
                c = self.grid.cellAt(col_num, row_num)
                if c.isBlock:
                    solution += '.'
                    fill += '.'
                    markup += b'\x00'
                    rebus_board.append(0)
                elif len(c.solution) == 1:
                    solution += c.solution
                    fill += '-'
                    markup += b'\x80' if (col_num, row_num) in circled else b'\x00'
                    rebus_board.append(0)
                else:
                    solution += c.solution[0]
                    fill += '-'
                    rebus_board.append(rebus_index + 1)
                    rebus_table += '{:2d}:{};'.format(rebus_index, c.solution)
                    rebus_index += 1
                #END if/else
            #END for col_num
        #END for row_num
                    
        pz.solution = solution
        pz.fill = fill
        
        # Clues
        # there *must* be an "across" and "down" here, else we throw an exception
        all_clues = []
        num_dirs_found = 0
        for clue_list in self.clues:
            if clue_list['title'].lower() == 'across':
                num_dirs_found += 1
                for c in clue_list['clues']:
                    setattr(c, 'dir', 0)
                    all_clues.append(c)
            if clue_list['title'].lower() == 'down':
                num_dirs_found += 1
                for c in clue_list['clues']:
                    setattr(c, 'dir', 1)
                    all_clues.append(c)
                
        if num_dirs_found != 2:
            raise(BaseException('Proper clue lists not found'))
            
        weirdass_puz_clue_sorting = sorted(all_clues, key=lambda c: (int(c.number), c.dir))
        
        clues = [c.clue for c in weirdass_puz_clue_sorting]
        
        normalized_clues = [unidecode_fxn(clue) for clue in clues]
        pz.clues.extend(normalized_clues)

        has_markup = b'\x80' in markup
        has_rebus = any(rebus_board)

        if has_markup:
            pz.extensions[b'GEXT'] = markup
            pz._extensions_order.append(b'GEXT')
            pz.markup()

        if has_rebus:
            pz.extensions[b'GRBS'] = bytes(rebus_board)
            pz.extensions[b'RTBL'] = rebus_table.encode(puz.ENCODING)
            pz._extensions_order.extend([b'GRBS', b'RTBL'])
            pz.rebus()
        
        # Save the file
        pz.save(filename)
        
        return
    #END toPuz()
        
    def toIPuz(self, filename):
        """Write an iPuz file"""
        d = {}
        # Metadata first
        d["version"] = "http://ipuz.org/v1"
        ipuzkind = f"http://ipuz.org/{self.metadata.kind}#1"
        d['kind'] = [ipuzkind]
        for a in ('author', 'title', 'copyright', 'notes'):
            d[a] = getattr(self.metadata, a, '')
        # dimensions
        d['dimensions'] = {"width": self.grid.width, "height": self.grid.height}
        # we explicitly define "block" and "empty"
        BLOCK, EMPTY = '#', '_'
        d['block'] = BLOCK; d['empty'] = EMPTY
        # puzzle and solution
        puzzle, solution = [], []
        for y in range(self.grid.height):
            row, solrow = [], []
            for x in range(self.grid.width):
                c = self.grid.cellAt(x, y)
                if c.isBlock:
                    row.append(BLOCK)
                elif c.isEmpty:
                    row.append(None)
                else:
                    num = c.number or EMPTY
                    row.append({"cell": num, "style": c.style})
                solrow.append(c.solution)
            #END for x
            puzzle.append(row)
            solution.append(solrow)
        #END for y
        d['puzzle'] = puzzle
        # add a solution only if there is one
        solset = set(list(itertools.chain(*solution)))
        if not solset.issubset(set([BLOCK, None])):
            d['solution'] = solution

        # Take care of clues, remembering that they are 1-indexed
        clues = OrderedDict()
        for c1 in self.clues:
            c1_arr = []
            for c2 in c1['clues']:
                thisClue = {"clue": c2.clue, "number": c2.number}
                thisCells = []
                for c3 in c2.cells:
                    thisCells.append([c3[0]+1, c3[1]+1])
                thisClue["cells"] = thisCells
                c1_arr.append(thisClue)
            #END for c2
            clues[c1['title']] = c1_arr
        #END for c1
        d['clues'] = clues

        # write the file
        with open(filename, 'w') as fid:
            json.dump(d, fid)
    #END toIPuz

    def fromDict(self, d1):
        """
        our file_types folder creates standard dictionaries
        this method creates a Puzzle instance from the dictionary
        """
        # Metadata
        md = d1['metadata']
        metadata = MetaData(md['kind'])
        metadata.title = md.get('title')
        metadata.author = md.get('author')
        metadata.copyright = md.get('copyright')
        metadata.notes = md.get('notes')
        metadata.width = md.get('width')
        metadata.height = md.get('height')

        # Grid
        cells = []
        for c in d1['grid']:
            cell = Cell(c['x'], c['y'], solution=c.get('solution')
                , value=c.get('value'), number=c.get('number')
                , isBlock=c.get('isBlock'), isEmpty=c.get('isEmpty'), style=c.get('style', {}))
            cells.append(cell)
        #END for c
        grid = Grid(cells)

        # Clues
        clues = []
        # We use these if we don't have explicit clue cell values
        if d1['metadata'].get('noClueCells'):
            cellLists = (grid.acrossEntries(), grid.downEntries())
        else:
            cellLists = ({}, {})
        for i, cluelist in enumerate(d1['clues']):
            title = cluelist['title']
            clues1 = cluelist['clues']
            thisClues = []
            for j, clue in enumerate(clues1):
                number = clue.get('number')
                # Infer cell locations if they're not given
                cells = clue.get('cells', cellLists[i].get(number, {}).get('cells'))
                c = Clue(clue.get('clue'), cells, number=number)
                thisClues.append(c)
            #END for clues1
            clues.append({'title': title, 'clues': thisClues})
        #END for cluelists
        return Puzzle(metadata=metadata, grid=grid, clues=clues)
    #END fromIPuz()

    def fromIPuz(self, puzFile):
        ipz = ipuz.read_ipuzfile(puzFile)
        return Puzzle().fromDict(ipz)
    #END fromIPuz()

    def fromCFP(self, puzFile):
        cfpdata = cfp.read_cfpfile(puzFile)
        return Puzzle().fromDict(cfpdata)
    #END fromCFP()

    def fromJPZ(self, puzFile):
        jpzdata = jpz.read_jpzfile(puzFile)
        return Puzzle().fromDict(jpzdata)
    #END fromJPZ()

    def fromAmuseLabs(self, s):
        data = amuselabs.read_amuselabs_data(s)
        return Puzzle().fromDict(data)
    #END fromAmuseLabs()
